import logging
from typing import Dict, List, TypeVar, Iterator, Type, Tuple
from typing_inspect import is_optional_type
from dataclasses import fields, is_dataclass, astuple as _astuple, asdict as _asdict

logger = logging.getLogger('serde')

JsonValue = TypeVar('JsonValue', str, int, float, bool, Dict, List)

T = TypeVar('T')

SE_NAME = '__serde_serialize__'

FROM_TUPLE = '__serde_from_tuple__'

FROM_DICT = '__serde_from_dict__'


class SerdeError(TypeError):
    """
    Serde error class.
    """


def gen(code: str, globals: Dict = None, locals: Dict = None):
    """
    Customized `exec` function.
    """
    logger.debug(code)
    exec(code, globals, locals)


def iter_types(cls: Type) -> Iterator[Type]:
    """
    Iterate field types recursively.
    """
    if is_dataclass(cls):
        yield cls
        for f in fields(cls):
            yield from iter_types(f.type)
    elif is_optional_type(cls):
        yield from iter_types(cls.__args__[0])
    elif issubclass(cls, List):
        yield from iter_types(cls.__args__[0])
    elif issubclass(cls, Tuple):
        for arg in cls.__args__:
            yield from iter_types(arg)
    elif issubclass(cls, Dict):
        yield from iter_types(cls.__args__[0])
        yield from iter_types(cls.__args__[1])
    else:
        yield cls


def astuple(v):
    """
    Convert decoded JSON `dict` to `tuple`.
    """
    if is_dataclass(v):
        return _astuple(v)
    elif isinstance(v, Dict):
        return tuple(astuple(e) for e in v.values())
    elif isinstance(v, (Tuple, List)):
        return tuple(astuple(e) for e in v)
    else:
        return v


def asdict(v):
    return _asdict(v)



from typing import Any, Tuple, Type

import msgpack
from dataclasses import astuple

from .core import T, logger
from .de import Deserializer, from_obj
from .se import Serializer


class MsgPackSerializer(Serializer):
    def serialize(self, obj: Any, **opts) -> str:
        return msgpack.packb(astuple(obj), **opts)


class MsgPackDeserializer(Deserializer):
    def deserialize(self, s: bytes, **opts) -> Tuple:
        unp = msgpack.unpackb(s, raw=False, use_list=False, **opts)
        logger.debug('unpack from msgpack: {unp}')
        return unp


def to_msgpack(obj: Any, serializer=MsgPackSerializer, **opts) -> bytes:
    return obj.__serde_serialize__(serializer, **opts)


def from_msgpack(c: Type[T], s: str, de: Type[Deserializer] = MsgPackDeserializer, **opts) -> Type[T]:
    return from_obj(c, s, de)



import json
from typing import Any, Dict, Tuple, Type

from dataclasses import asdict

from .core import JsonValue, T
from .de import Deserializer, from_obj
from .se import Serializer, is_serializable


class JsonSerializer(Serializer):
    def serialize(self, obj: Any) -> str:
        return json.dumps(asdict(obj))


class JsonDeserializer(Deserializer):
    def deserialize(self, s: str) -> Dict[str, JsonValue]:
        return json.loads(s)


def to_json(obj: Any, serializer=JsonSerializer) -> str:
    if is_serializable(obj):
        return obj.__serde_serialize__(serializer)
    else:
        return json.dumps(obj)


def from_json(c: Type[T], s: str, de: Type[Deserializer] = JsonDeserializer) -> T:
    return from_obj(c, s, de)



import functools
from typing import Any, Type

from .core import SE_NAME, T


class Serializer:
    def serialize(self, obj: Any) -> Any:
        pass


def serialize(_cls: Type[T]) -> Type[T]:
    """
    `serialize` decorator.
    """

    @functools.wraps(_cls)
    def wrap(cls: Type[T]) -> Type[T]:
        def serialize(self, ser, **opts) -> None:
            return ser().serialize(self, **opts)

        setattr(cls, SE_NAME, serialize)
        return cls

    return wrap(_cls)


def is_serializable(instance_or_class: Any) -> bool:
    """
    Test if `instance_or_class` is serializable.
    """
    return hasattr(instance_or_class, '__serde_serialize__')



import pathlib
import re
import sys
from codecs import open
from setuptools import setup, find_packages

root = pathlib.Path(__file__).parent.absolute()

with open('serde/__init__.py', 'r', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

with open('README.md', 'r', encoding='utf8') as f:
    readme = f.read()

setup_requires = [
    'pytest-runner',
]

# Installs dataclasses from PyPI for python < 3.7
if sys.version_info < (3, 7):
    requires = [
        'dataclasses',
    ]
# Native dataclasses support for python >= 3.7
else:
    requires = [
    ]

tests_require = [
    'coverage',
    'pytest',
    'pytest-cov',
    'pytest-flake8',
    'mypy',
    'flake8',
]

setup(
    name='pyserde',
    version=version,
    description='Dataclass based serialization library inspired by serde.',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='yukinarit',
    author_email='yukinarit84@gmail.com',
    url='https://github.com/yukinarit/pyserde',
    packages=find_packages(exclude=['test_serde', 'bench']),
    python_requires=">=3.6",
    setup_requires=setup_requires,
    install_requires=requires,
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
    },
    license='MIT',
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from typing import Any, Dict, Tuple, List, Type
from typing_inspect import is_optional_type
from dataclasses import fields, is_dataclass

from .core import SerdeError, FROM_DICT, FROM_TUPLE, T, gen, iter_types


def from_any(cls, o):
    if not is_deserializable(cls):
        raise SerdeError('`cls` must be deserializable.')

    if isinstance(o, (List, Tuple)):
        return cls.__serde_from_tuple__(o)
    elif isinstance(o, Dict):
        return cls.__serde_from_dict__(o)
    elif isinstance(o, cls) or o is None:
        return o
    else:
        raise SerdeError(f'`o` must be either List, Tuple, Dict or cls but {type(o)}.')


def from_tuple(cls) -> str:
    """
    Generate function to deserialize from tuple.
    """
    params = []
    for i, f in enumerate(fields(cls)):
        params.append(f'{f.name}=' + from_value(f.type, f'data[{i}]'))
    return ', '.join(params)


def from_dict(cls) -> str:
    """
    Generate function to deserialize from dict.
    """
    params = []
    for f in fields(cls):
        params.append(f'{f.name}=' + from_value(f.type, f'data["{f.name}"]'))
    return ', '.join(params)


def from_value(typ: Type, varname: str) -> str:
    """
    Generate function to deserialize from value.
    """
    if is_deserializable(typ):
        nested = f'{typ.__name__}'
        s = f"from_any({nested}, {varname})"
    elif is_optional_type(typ):
        element_typ = typ.__args__[0]
        s = f"{from_value(element_typ, varname)}"
    elif issubclass(typ, List):
        element_typ = typ.__args__[0]
        s = f"[{from_value(element_typ, 'd')} for d in {varname}]"
    elif issubclass(typ, Dict):
        key_typ = typ.__args__[0]
        value_typ = typ.__args__[1]
        s = (f"{{ {from_value(key_typ, 'k')}: {from_value(value_typ, 'v')} "
             f"for k, v in {varname}.items() }}")
    elif issubclass(typ, Tuple):
        elements = [from_value(arg, varname + f'[{i}]') + ', ' for i, arg in enumerate(typ.__args__)]
        s = f"({''.join(elements)})"
    else:
        s = varname
    return s


def gen_from_function(cls: Type[T], funcname: str, params: str) -> Type[T]:
    """
    Generate function to deserialize from tuple.
    """
    body = (f'def {funcname}(data):\n return cls({params})')

    globals: Dict[str, Any] = dict(cls=cls)

    # Collect fields to be used in the scope of exec.
    for typ in iter_types(cls):
        if is_dataclass(typ):
            globals[typ.__name__] = typ
    globals['from_any'] = from_any

    gen(body, globals)
    setattr(cls, funcname, staticmethod(globals[funcname]))

    return cls


def is_deserializable(instance_or_class: Any) -> bool:
    """
    Test if `instance_or_class` is deserializable.
    """
    return hasattr(instance_or_class, FROM_TUPLE) or hasattr(instance_or_class, FROM_DICT)


class Deserializer:
    """
    Deserializer base class.
    """
    def deserialize(self, obj, **opts):
        return obj


def deserialize(_cls=None, rename_all: bool = False) -> Type:
    """
    `deserialize` decorator. A dataclass with this decorator can be
    deserialized into an object from various data interchange format
    such as JSON and MsgPack.
    """
    def wrap(cls) -> Type:
        cls = gen_from_function(cls, FROM_TUPLE, from_tuple(cls))
        cls = gen_from_function(cls, FROM_DICT, from_dict(cls))
        return cls

    if _cls is None:
        wrap

    return wrap(_cls)


def from_obj(c: Type[T], o: Any, de: Type[Deserializer] = None, **opts) -> T:
    if de:
        o = de().deserialize(o, **opts)
    if o is None:
        return None
    if is_deserializable(c):
        return from_any(c, o)
    elif is_optional_type(c):
        return from_obj(c.__args__[0], o)
    elif issubclass(c, List):
        return [from_obj(c.__args__[0], e) for e in o]
    elif issubclass(c, Tuple):
        return tuple(from_obj(c.__args__[i], e) for i, e in enumerate(o))
    elif issubclass(c, Dict):
        return {from_obj(c.__args__[0], k): from_obj(c.__args__[1], v) for k, v in o.items()}
    else:
        return o
